/** Java 1.3 AST Recognizer Grammar
 *
 * Author: (see java.g preamble)
 *
 * This grammar is in the PUBLIC DOMAIN
 */
class JavaRefsTreeParser extends JavaTreeParser;

options {
	importVocab = Java;
}

{
    java.util.Stack scopeStack = new java.util.Stack();
    CodeProject project = null;
    MethodScope currentMethodScope = null;

    public void setProject(CodeProject project) {
        this.project = project;
    }

    protected String fileName = null;
    public void setFileName(String f) {
        fileName = f;
    }
}

compilationUnit
{
    FileScope fs = project.getFileScope(fileName);
    scopeStack.push(fs);
}
    :
        (packageDefinition)?
		(importDefinition)*
		(typeDefinition)*
	;

importDefinition
	:	#( IMPORT packageOrClassIdentifier )
	;

/** only called from importDefinition */
packageOrClassIdentifier
{
    String r = null;
    Scope scope = null;
}
	:	IDENT
        {
        FileScope currentScope = (FileScope)scopeStack.peek();
        ClassScope cs=project.getClassScope(#IDENT.getText());
        currentScope.addImport(scope);
        // System.out.println("import class "+scope);
        }
	|	#( DOT r=identifier
           (    STAR
                {
                FileScope currentScope = (FileScope)scopeStack.peek();
                scope = project.getPackageScope(r);
                currentScope.addImport(scope);
                // System.out.println("import "+scope);
                }
           |    IDENT
                {
                String id=r+"."+#IDENT.getText();
                FileScope currentScope = (FileScope)scopeStack.peek();
                ClassScope cs=project.getClassScope(id);
                currentScope.addImport(cs);
                // System.out.println("import class "+cs);
                }
           )
         )
	;

typeDefinition
{
    ClassScope cs = null;
}
:	#(  CLASS_DEF modifiers IDENT
            {
            Scope currentScope = (Scope)scopeStack.peek();
            String pName = currentScope.getFullyQualifiedName();
            cs = project.getClassScope(pName+"."+#IDENT.getText());
            }
            extendsClause[cs] implementsClause[cs]
            {scopeStack.push(cs);}
            objBlock
            {scopeStack.pop();}
         )
	|	#(  INTERFACE_DEF modifiers IDENT
            {System.out.println("interface "+#IDENT.getText());}
            {
            Scope currentScope = (Scope)scopeStack.peek();
            String pName = currentScope.getFullyQualifiedName();
            cs = project.getClassScope(pName+"."+#IDENT.getText());
            }
            extendsClause[cs]
            {scopeStack.push(cs);}
            interfaceBlock
            {scopeStack.pop();}
         )
	;

extendsClause[ClassScope def]
{
    String id = null;
}
	:	#(  EXTENDS_CLAUSE
            (   id=identifier
                {
                Scope currentScope = (Scope)scopeStack.peek();
                ClassScope cs = currentScope.resolveClass(id);
                def.addSuperClass(cs);
                }
            )*
         )
         {
         if ( def!=null && def.getSuperClasses()==null ) {
            ClassScope o = project.getClassScope("java.lang.Object");
            def.addSuperClass(o);
         }
         }
	;

implementsClause[ClassScope def]
{
    String id = null;
}
	:	#(  IMPLEMENTS_CLAUSE
            (   id=identifier
                {
                Scope currentScope = (Scope)scopeStack.peek();
                ClassScope cs = currentScope.resolveClass(id);
                def.addImplementedInterface(cs);
                }
            )*
         )
	;


interfaceBlock
	:	#(	OBJBLOCK
			(	methodDecl
			|	variableDef
			)*
		)
	;

typeSpec returns [ClassScope t=null]
	:	#(TYPE t=typeSpecArray)
	;

typeSpecArray returns [ClassScope t=null]
	:	#( ARRAY_DECLARATOR t=typeSpecArray )
	|	t=type
	;

type returns [ClassScope t=null]
{
    String r=null;
}
    :	r=identifier
        {
        Scope currentScope = (Scope)scopeStack.peek();
        t = currentScope.resolveClass(r);
        }
	|	builtInType
	;

objBlock
	:	#(	OBJBLOCK
			(	ctorDef
			|	methodDef
			|	fieldDef
			|	typeDefinition
			|	#(STATIC_INIT slist)
			|	#(INSTANCE_INIT slist)
			)*
		)
	;

fieldDef
{
    ClassScope t = null;
}
	:	#(VARIABLE_DEF modifiers t=typeSpec fieldDeclarator[t] varInitializer)
	;

fieldDeclarator[ClassScope t]
	:	IDENT
        {
            // set types of all fields now that we can resolve refs
            ClassScope currentScope = (ClassScope)scopeStack.peek();
            FieldVariable f = currentScope.getField(#IDENT.getText());
            f.setType(t);
        }
	|	LBRACK fieldDeclarator[t]
	;

methodDef
{
    ClassScope retType = null; // either class name or we don't care
}
	:	#(METHOD_DEF modifiers retType=typeSpec methodHead[retType] (slist)?)
        {scopeStack.pop(); scopeStack.pop();}
	;

ctorDef
	:	#(CTOR_DEF modifiers methodHead[null] (slist)?)
        {scopeStack.pop(); scopeStack.pop();}
	;

methodDecl
{
    ClassScope retType = null; // either class name or we don't care
}
	:	#(METHOD_DEF modifiers retType=typeSpec methodHead[retType])
	;

methodHead[ClassScope retType]
	:	IDENT
        {
            Scope currentScope = (Scope)scopeStack.peek();
            String pName = currentScope.getFullyQualifiedName();
            MethodScope ms = project.getMethodScope(pName+"."+#IDENT.getText());
            if ( ms!=null ) {
                scopeStack.push(ms);
                ParameterScope ps =
                    new ParameterScope(project,"[parameters]",ms);
                scopeStack.push(ps);
                ms.setReturnType(retType);
                ms.setParameterScope(ps);
                currentMethodScope = ms;
            }
            else {
                System.out.println("# cannot find method "+pName+"."+#IDENT.getText());
            }
        }
        #( PARAMETERS (parameterDef)* ) (throwsClause)?
	;

slist
	:	#(  SLIST
            {
            // local {...} blocks are not kept in the MethodScope.
            // They are pushed on temporarily for resolving variables
            // but are not stored for use later.
            Scope currentScope = (Scope)scopeStack.peek();
            CodeBlockScope cbs =
                new CodeBlockScope(project,"[slist]",currentScope);
            scopeStack.push(cbs);
            }
            (stat)*
            {Scope s = (Scope)scopeStack.pop(); s.dump(); }
         )
	;

variableDef
{
    ClassScope t = null;
}
	:	#(VARIABLE_DEF modifiers t=typeSpec variableDeclarator[t] varInitializer)
	;

variableDeclarator[ClassScope t]
	:	IDENT
        {
            Scope currentScope = (Scope)scopeStack.peek();
            LocalVariable local =
                new LocalVariable(project,#IDENT.getText(),currentScope);
            local.setType(t);
            currentScope.addScopeMember(local);
        }
	|	LBRACK variableDeclarator[t]
	;

parameterDef
{
    ClassScope t = null;
}
	:	#(PARAMETER_DEF modifiers t=typeSpec IDENT )
        {
            Scope currentScope = (Scope)scopeStack.peek();
            LocalVariable local =
                new LocalVariable(project,#IDENT.getText(),currentScope);
            local.setType(t);
            currentScope.addScopeMember(local);
        }
	;


identifier returns [String id=null]
{
    String r = null;
}
	:	IDENT {id=#IDENT.getText();}
	|	#( DOT r=identifier IDENT ) {id=r+"."+#IDENT.getText();}
	;

primaryExpression returns [ClassScope t=null]
    :   IDENT
    |   #(	DOT
			(	expr
				(	IDENT
				|	arrayIndex
				|	"this"
				|	"class"
				|	#( "new" IDENT elist )
				|   "super"
				)
			|	#(ARRAY_DECLARATOR typeSpecArray)
			|	builtInType ("class")?
			)
		)
	|	arrayIndex
	|	methodCallExpression
	|	#(TYPECAST typeSpec expr)
	|   newExpression
	|   constant
    |   "super"
    |   "true"
    |   "false"
    |   "this"
    |   "null"
	|	typeSpec // type name used with instanceof
	;

/** Isolate method calls since the context information allows us
 *  to make a simpler, more explicit grammar.
 */
methodCallExpression returns [ClassScope t=null]
{
    ClassScope objOrTypeScope = null;
}
    :   #(  METHOD_CALL
            (   IDENT
                {
                Scope currentScope = (Scope)scopeStack.peek();
                MethodScope ms = currentScope.resolveMethod(#IDENT.getText());
                if ( ms!=null ) {
                    project.refMethod(currentMethodScope, ms, #methodCallExpression);
                    t = ms.getReturnType();
                }
                }
            |	#( DOT objOrTypeScope=objectOrClassExpression IDENT )
                {
                if ( objOrTypeScope!=null ) {
                    MethodScope ms = objOrTypeScope.resolveMethod(#IDENT.getText());
                    if ( ms!=null ) {
                        project.refMethod(currentMethodScope, ms, #methodCallExpression);
                        t = ms.getReturnType();
                    }
                }
                }
            )
            elist
         )
    ;

objectOrClassExpression returns [ClassScope t=null]
{
    ClassScope objOrTypeScope = null;
    String id = null;
}
    :   t=methodCallExpression
    |   #(TYPECAST t=typeSpec expr)
    |   t=newExpression
        {System.out.println("new expr resolved to "+
         (t!=null?t.getFullyQualifiedName():"null"));}
    |   STRING_LITERAL {t=project.getClassScope("java.lang.String");}
    |   "super"
        {
        ClassScope cs = (ClassScope)currentMethodScope.getParent();
        t=cs.getSuperClass();
        System.out.println("super resolved to "+t.getFullyQualifiedName());
        }
    |   "this"
        {
        t = (ClassScope)currentMethodScope.getParent(); // get class scope.
        System.out.println("this resolved to "+t.getFullyQualifiedName());
        }

    |	t=arrayIndexInMethodCallExpression

        // com.jguru.test.Main.getName();  package.class access
        // a.foo();     field access
        // a.b.c.foo(); field access
        // Main.foo();  class method access
        // java.lang.System.out.println(); package.class.field access

    |   (identifier)=> id=identifier
        {
        System.out.print("qname (class or field) ref "+id);
        Scope currentScope = (Scope)scopeStack.peek();
        t = project.resolveTypeOrField(currentScope, id);
        }

        /*
    |   // Used with Main.foo() or user.getName(); must be class or var/field
        IDENT
        {
        System.out.print("ID ref "+#IDENT.getText());
        Scope currentScope = (Scope)scopeStack.peek();
        Variable v = currentScope.resolveVariable(#IDENT.getText());
        if ( v!=null ) {
            System.out.print("var resolved to ");
            v.dump();
            t = v.getType();
        }
        }
        */

    |   #(  DOT
            (   objOrTypeScope=objectOrClassExpression
				(	IDENT
                    {
                    if ( objOrTypeScope!=null ) {
                        FieldVariable f = objOrTypeScope.resolveField(#IDENT.getText());
                        if ( f!=null ) {
                            t = f.getType();
                        }
                    }
                    }
				|	"this" {t=objOrTypeScope;}
				|	"class"{t=objOrTypeScope;}
				|	#( "new" IDENT elist )
                    {
                    Scope currentScope = (Scope)scopeStack.peek();
                    t = currentScope.resolveClass(#IDENT.getText());
                    }
				)
			|	builtInType "class"
            )
         )
    ;

newExpression returns [ClassScope t=null]
	:	#(	"new" t=type
			(	newArrayDeclarator (arrayInitializer)?
			|	elist (objBlock)?
			)
		)
	;

arrayName returns [ClassScope t=null]
    :   t=objectOrClassExpression
    |   #(QUESTION expr expr expr)	// trinary operator
        // ambiguous; ignore these
    ;

arrayIndexInMethodCallExpression returns [ClassScope t=null]
	:	#(INDEX_OP t=arrayName expression)
        {System.out.println("array ref of type "+t.getFullyQualifiedName());}
	;
